#!/usr/bin/perl
#
# Snare Audit Dispatcher for Linux
# (c) Copyright 2010 InterSect Alliance Pty Ltd
#
# This application will integrate into the native linux audit subsystem,
# and translate linux auditd log format data into something that can be
# parsed appropriately by apps that prefer one-line-per-event, such as the
# Snare audit daemon (for delivery to the Snare Server), or logwatch.
#
# INSTALLATION:
#
#	* Run the MakeTranslationTable program
#	  (optional, but highly recommended)
#         It will create /etc/snare-xlate.conf
#	* Copy SnareDispatchHelper to /usr/sbin/SnareDispatchHelper
#	* Save this file to /usr/sbin/SnareDispatcher.pl
#	* Add the following line to your /etc/auditd.conf file:
#		dispatcher = /usr/sbin/SnareDispatchHelper
#
# Issues:
#	* a0, a1, a2, a3 are practically useless, as string arguments are not
#	  supported. execve's are a particular problem.
#	* It is programatically difficult to determine how many 'lines' make
#         up an audit event. Some lines can be repeated, with slightly
#         different values.
#       * You can have multiple, identical tokens for an event (eg: two path=)
#	  we handle this by appending a count (after a colon).
#	* Event lines may be interleaved (ie: you might get two lines from
#	  event # 1000, then one line from event # 1001, then another line
#         from event # 1000).
#	* Some filename characters are translated into their HEX equivalents
#         which will make matching filenames difficult.
#           if (char == '"' || char < 0x21 || char > 0x7f)
#

use POSIX qw(strftime);
use Socket;
use Sys::Syslog qw ( :DEFAULT setlogsock );
use Digest::MD5 qw(md5 md5_hex);
use MIME::Base64;
use File::Basename;
use Fcntl;
use constant DEBUG => 0;

use constant SUCCESS => 1;
use constant FAILURE => 2;

use constant WATCH => 1;
use constant NEWWATCH => 2;
use constant IGNORE => 3;

############# User Configurable options.

$WEBSERVER="/usr/sbin/SnareWebServer.pl";
$CONFIGFILE="/etc/snare.conf";

# Note: These two values can be configured using the config file.

# Should we care about the criticality?
# If not, no worries - we can return as soon as we find a single match,
# which will speed things up a little.
$USECRITICALITY=1;

# Should we manage the audit configuration? Or do we leave it to the native audit subsystem?
$MANAGEAUDIT=1;

# Should we use regular expressions instead of wildcards?
$REGEX=0;

# Should we use watches instead of system-wide syscalls?
$USEWATCH=0;

$WEBSERVERALLOW=0;
$WEBSERVERPORT=6161;
$ACCESSKEY="";
$RESTRICT_IP="";

# Where are we sending events?
$SYSLOGDEST=0;

############# End user configurable options.

$VERSION="1.5.1";

$CONTINUE=1;

# Open our translation table.
$rc=open(XLATE,"/etc/snare-xlate.conf");
if($rc) {
	while($line=<XLATE>) {
		chomp($line);
		($num,$scall)=split(/:/,$line);
		$syscall{$num}=$scall;
		$nsyscall{$scall}=$num;
	}
	$syscall{-1}="login_auth";
	$syscall{-2}="login_start";
	$syscall{-3}="logout";
	$syscall{-4}="acct_mgmt";
	close(XLATE);
}


# Declare a UID/GID cache so we don't wind up getting into recursive
# audit event generation.
my %uidcache=();
my %unamecache=();
my %gidcache=();

# Event cache
%event=();

# Global filters array
%Filters=();
$FilterCount=0;
$WatchCount=0;
%WatchVal=();
%WatchExtra=();

$DNSTIMEOUT=600;

$FQDN=`hostname --fqdn`;
chomp($FQDN);
$FQDN = "UNKNOWN" if ($FQDN eq "");

# Set up some select() related variables.
$rin = '';
vec($rin,fileno(STDIN),1) = 1;

LoadConfig() or exit;

# Start up our web server (if allowed)
StartWeb();

# Set up our audit configuration based on our objectives.
SetAudit();

# Trap a few signals
$SIG{USR1}='Restart';
$SIG{USR2}='DumpEvents';
$SIG{PIPE}='Handler';
$SIG{CHLD}='IGNORE';
$SIG{HUP}='Handler';
$SIG{INT}='Handler';
$SIG{QUIT}='IGNORE';
$SIG{TERM}='Handler';

LogMsg("Snare for Linux $VERSION: Started and active.");
$DNSNOW=time();
$DNSLAST=$DNSNOW;

# Main reader loop.
while($CONTINUE>=1) {
	$nfound=select($rout=$rin,undef,undef,1);
	if($nfound > 0) {
		# Grab the version number
		$bytes=sysread(STDIN,$header,16);
		if($bytes==0) {
			if (DEBUG) { LogMsg("DEBUG: Failed to get header\n"); }
			$CONTINUE=0;
			last;
		} elsif ($bytes==undef) {
			# Probably interrupted by a signal
			next;
		}

		# Eventtype 1302 = PATH
		# Eventtype 1300 = SYSCALL
		# Eventtype 1307 = CWD
		($eventtype,$contentsize,$eventnum,$datetime)=unpack("IIII",$header);
		# 8k sanity check. Hopefully, we'll resynchronise.
		if($contentsize > 0 && $contentsize < 16384) {
			sysread(STDIN,$line,$contentsize);
		} else {
			if(DEBUG) { LogMsg("Content is $contentsize! Discarding"); }
			sysread(STDIN,$line,$contentsize);
			next;
		}
	} else {
		$CONTINUE++ if ($CONTINUE);
		if(DEBUG>=1) { LogMsg("NO DATA (timeout)! $CONTINUE"); }
		next;
	}


#DMM
	#SendEvent("LinuxTEST","$line\n");
	#print $OUTFILE "$line\n";
	#next;
	#chomp;
	if(DEBUG>=2) { LogMsg("\nLINE: $line"); }

	if(DEBUG>=1) { LogMsg("datetime: $datetime eventnum: $eventnum tail: $line"); }

	if($datetime==0 || !$line || !$eventnum) {
		# Not interested
		next;
	}

	if(DEBUG>=1) { LogMsg("DATE TIME = $datetime EVENTID = $eventnum eventtype = $eventtype"); }


	# Lets split this line up into element/content pairs.
	# First, lets check for msg strings and strip them out (to deal with them later)
	# msg is the only string that contains spaces
	if ($line =~ /msg='/) {
		if ($line =~ /(.* )msg='(.*?)'(.*)/) {		# Make gedit pretty: '
			$msg = $2;
			$line = $1.$3;
		}
	}
	if(DEBUG>=2) { LogMsg("LINE(-msg): $line"); }
	# Next, break apart by spaces that aren't inside inverted commas.
	#@elements=split(/\s+(?=(?:[^"]*"[^"]*")*[^"]*\z)/,$line);  # comment helper """
	@elements=split(/\s+/,$line);  # comment helper """
	#DMM - for encoded text
	# $txt =~ s/([a-fA-F0-9][a-fA-F0-9])/chr(hex($1))/eg
	%event=();
	if(DEBUG>=2) { LogMsg("Elements contains " . @elements . " elements"); }
	foreach (@elements) {
		# Then, by the equals sign.
		($key,$val)=split(/=/,$_,2);
		next if ($val =~ /^$/);
		$val =~ s/"//g;		# Make gedit pretty: "
		#$val =~ s/^"(.*)"$/$1/;
		$count=0;
#			#push(@{$event{$key}},$val);//slighly slower
		while(defined($event{$key . ($count==0?"":":$count")})) {
			$count++;
		}

		$event{$key . ($count==0?"":":$count")}=$val;
		if(DEBUG>=4) { LogMsg("ASSIGN: $key:$count = $val"); }
	}
	# Now, do the same thing for the msg string, but add a few special clauses
	# Strip out any brackets
	# structure: msg='(messge|key=val): (message|key=val) \((key=val[, ]*)*\)'
	if (defined $msg) {
		if(DEBUG>=4) { LogMsg("MESSAGE: $msg"); }
		undef $key;
		#split the message again
		($hdr,$tail,$null) = split(/\((.*?)\)/,$msg);  # comment helper """
		if(DEBUG>=4) { LogMsg("MESSAGE HDR: $hdr"); }
		if(DEBUG>=4) { LogMsg("MESSAGE HDR: $tail"); }
		@elements=split(/[,]?[:\s]+(?=(?:[^"]*"[^"]*")*[^"]*\z)/,$hdr);  # comment helper """
		@elements2=split(/,?\s(?=(?:[^"]*"[^"]*")*[^"]*\z)/,$tail);  # comment helper """
		if(DEBUG>=2) { LogMsg("Elements contains " . @elements . " elements"); }
		if(DEBUG>=2) { LogMsg("Elements2 contains " . @elements2 . " elements"); }
		foreach $element (@elements, @elements2) {
			if($element=~/=/) {
				# Then, by the equals sign.
				($key,$val)=split(/=/,$element,2);
				$val =~ s/^"//;		# Make gedit pretty: "
				$val =~ s/"$//;		# Make gedit pretty: "
				$count=0;
				while(defined($event{$key . ($count==0?"":":$count")})) {
					$count++;
				}

				$event{$key . ($count==0?"":":$count")}=$val;
				#$event{$key}=$val;
				if(DEBUG>=4) { LogMsg("ASSIGN: $key.$count = $val"); }
			} else {
				if (!defined($key)) {
					$key = "msg";
					$event{$key} = $element;
					next;
				}
				# add the term to the previous key
				$event{$key . ($count==0?"":":$count")} .= " $element";
				if(DEBUG>=4) { LogMsg("ASSIGN: $key.$count = $element"); }
			}
		}
		undef $msg;
	}

	# Ok, we now have an array of key/value pairs.

	#always use the audit datetime
	@ltime=localtime($datetime);
	$sdatetime=strftime("%Y%m%d %T",@ltime);
	$event{"datetime"}=$sdatetime;

	# Eventtype 1302 = "PATH"
	# Eventtype 1300 = SYSCALL
	# Eventtype 1112 = "USER_LOGIN"
	# This should tell us when the end of a record comes through.
	if($eventtype == 1300) {
		# We have all the data we need! Send this event out the door.
		if(DEBUG>=1) { LogMsg("Sending out Event $eventnum"); }
		PrepEvent();
	} elsif($eventtype == 1112 || $eventtype == 1100 || $eventtype == 1105 || $eventtype == 1106 || $eventtype == 1108) {
#1112,1100,1105 - authentication
#1106 - USER_END
#1108 - USER_CHAUTHTOK
#1104 - ??
		if(DEBUG>=1) { LogMsg("User Login event detected, formatting and sending"); }
		if(DEBUG>=1) { LogMsg("Sending out Event $eventnum"); }
		# set up some of the missing variables
		$event{"gid"}=0;
		$event{"exit"}=0;
		if (defined($event{"terminal"})) {
			$event{"comm"}=$event{"terminal"};
		} else {
			$event{"comm"}="-";
		}
		$event{"success"} = "no";
		if (defined($event{"res"}) && $event{"res"} =~ /success/i) {
			$event{"success"} = "yes";
		} elsif (defined($event{"result"}) && $event{"result"} =~ /Success/i) {
			$event{"success"} = "yes";
		} elsif (!defined($event{"res"}) && !defined($event{"result"}) && $event{"msg"} !~ /fail/i) {
			$event{"success"} = "yes";
		} else {
			$event{"success"} = "no";
		}
		# login/logout specifics
		if ($eventtype == 1106) {
			#USER_END
			$event{"syscall"}=-3;
		} elsif ($eventtype == 1108) {
			# USER_CHAUTHTOK
			$event{"syscall"}=-4;
		} else {
			#USER_AUTH, USER_START
			if ($eventtype == 1105) {
				$event{"syscall"}=-2;
			} else {
				$event{"syscall"}=-1;
			}
		}
		if (!defined($event{"msg"})) {
			$event{"msg"} = $event{"op"};
		} else {
			$event{"msg"} .= " " . $event{"op"};
		}
		delete $event{"op"};
		delete $event{"res"};
		PrepEvent();
	}
	undef $event;
}

if (DEBUG) { LogMsg("DEBUG: Exit requested\n"); }

# disable audit
`/sbin/auditctl -e 0`;

if($webpid) {
	# send sigquit, which tells the web server thread to terminate cleanly.
	LogMsg("Sending signal 3 (Quit) to webpid $webpid\n");
	kill(3,$webpid);
}

# Flush our buffers
CloseOutputs();

exit;

sub PrepEvent {
	# If we are sent a particular event number, flush it out.
		#if(DEBUG) { LogMsg("DEBUG: Eventnum is $eventnum"); }

	# If the event doesn't satisfy the basic requirements, dump it
	if(!defined($event{"syscall"}) ||
	   !defined($event{"datetime"}) ||
	   !defined($event{"uid"}) ||
	   !defined($event{"gid"}) ||
	   !defined($event{"pid"}) ||
	   !defined($event{"comm"}) ||
	   !defined($event{"exit"}) ||
	   !defined($event{"success"})) {
		if(DEBUG) {
			LogMsg(":::::::: Event does not contain all required data");
			@elements=keys %event;
			@values=values %event;
			LogMsg(":::::::: [@elements] [@values]");
		}

		return;
	}


	# Lets construct a token/data array in the order we wish to send things out, with
	# appropriate data extrapolated.
	my %tokens=();
	my @tokenlist=();

	$syscallname=$syscall{$event{"syscall"}};
	if(!$syscallname) {
		# Fall back to the number.
		$syscallname=$event{"syscall"};
	}
	delete $event{"syscall"};
	$tokens{"event"}="$syscallname," . $event{"datetime"};
	push(@tokenlist,"event");
	delete $event{"datetime"};
	if ($event{"uid"} == -1) {
		$uidname=$event{"acct"};
	} else {
		$uidname=getuid($event{"uid"});
	}
	$tokens{"uid"}=$event{"uid"} . ($uidname?",$uidname":"");
	push(@tokenlist,"uid");
	delete $event{"uid"};
	if (defined($event{"id"})) {
		$idname=getuid($event{"id"},($syscallname eq "acct_mgmt"?1:0));
		if (!defined($event{"acct"})) {
			$event{"acct"} = $idname;
		}
	}
	$tokens{"id"}=$event{"id"} . ($idname?",$idname":"");
	push(@tokenlist,"id");
	delete $event{"id"};
	$gidname=getgid($event{"gid"});
	$tokens{"gid"}=$event{"gid"} . ($uidname?",$gidname":"");
	push(@tokenlist,"gid");
	delete $event{"gid"};
	$euidname=getuid($event{"euid"});
	$tokens{"euid"}=$event{"euid"} . ($uidname?",$euidname":"");
	push(@tokenlist,"euid");
	delete $event{"euid"};
	$egidname=getgid($event{"egid"});
	$tokens{"egid"}=$event{"egid"} . ($uidname?",$egidname":"");
	push(@tokenlist,"egid");
	delete $event{"egid"};
	$tokens{"process"}=$event{"pid"} . "," . $event{"comm"};
	push(@tokenlist,"process");
	delete $event{"pid"};
	delete $event{"comm"};
	$tokens{"return"}=$event{"exit"} . "," . $event{"success"};
	push(@tokenlist,"return");
	delete $event{"exit"};
	delete $event{"success"};


	foreach $key (sort keys %event) {
		# Resolve names
		if($key =~ /^[eosa]uid(:[0-9]+)*$/ || $key =~ /^fsuid(:[0-9]+)*$/) {
			$eventstring = $event{$key};
			$name=getuid($event{$key});
			if($name) {
				$eventstring .= ",$name";
			}
		} elsif($key =~ /^[eosa]gid(:[0-9]+)*$/ || $key =~ /^fsgid(:[0-9]+)*$/) {
			$eventstring = $event{$key};
			$name=getgid($event{$key});
			if($name) {
				$eventstring .= ",$name";
			}
		} elsif($key =~ /^name(:[0-9]+)*$/ || $key =~ /^exe(:[0-9]+)*$/) {
			$tpath=$event{$key};
			#DMM - for encoded text
			# $txt =~ s/([a-fA-F0-9][a-fA-F0-9])/chr(hex($1))/eg
			if($tpath =~ /^\// || !$event{"cwd"}) {
				$path = resolve_path($tpath);
			} else {
				$path = resolve_path($event{"cwd"} . "/" . $tpath);
			}
			$eventstring = $path;
		} else {
			$eventstring = $event{$key};
		}
		$tokens{$key}=$eventstring;
		push(@tokenlist,$key);
	}

	# Format of the filters:
	#$Filters{$ObjectiveCount}{$key}{$keymatch}{$ElementCount}{$AlternativeCount}=$alternative;
	#$Filters{0}{event}{0}{0}{0}="execve";
	$EventMatched=0;
	$EventCrit=0;
	FilterLOOP: for ($objectivecount=0; $objectivecount < $FilterCount;$objectivecount++) {
		$ObjectiveMatch=0;
		if(DEBUG>=3) { LogMsg("LOOP: objcount $objectivecount\n"); }
		foreach $key (keys(%{$Filters{$objectivecount}})) {
			if(DEBUG>=3) { LogMsg("LOOP: key $key\n"); }
			if($tokens{$key}) {
				foreach $keymatch (keys(%{$Filters{$objectivecount}{$key}})) {
					if(DEBUG>=3) { LogMsg("LOOP: keymatch $keymatch\n"); }
					foreach $elementcount (keys(%{$Filters{$objectivecount}{$key}{$keymatch}})) {
						if(DEBUG>=3) { LogMsg("LOOP: elementcount $elementcount\n"); }
						@elements=split(/,/,$tokens{$key});
						if(DEBUG>=3) { LogMsg("Key: $key - ELEMENTS: @elements\n"); }
						$result=0;
						foreach $alternativecount (keys(%{$Filters{$objectivecount}{$key}{$keymatch}{$elementcount}})) {
							if(DEBUG>=3) { LogMsg("LOOP: altcount $alternativecount\n"); }
							$match=$Filters{$objectivecount}{$key}{$keymatch}{$elementcount}{$alternativecount};
							$element=$elements[$elementcount];
							$negate=$FilterTypes{$objectivecount}{$key}{$keymatch};
							if(DEBUG>=2) { LogMsg("Match element $element against $match (negate is $negate)\n"); }

							# Special case:
							if($match eq "*") {
								$result=1;
							} else {
								@strings=split(/,/,$element);
								@matches=split(/,/,$match);
								$count=0;
								foreach $tmatch (@matches) {
									if ($REGEX == 1) {
										$regex=$tmatch;
										$regex =~ s/[^\\]\//\\\//g;
									} else {
										#$regex=wildcard($tmatch);
										my $re = quotemeta $tmatch;
										$re =~ s/\\\*/(.*)/g;
										$regex = "^" . $re . "\$";
									}
									if(DEBUG>=1) { LogMsg("Match: $regex against " . $strings[$count]); }
									if($strings[$count] =~ /$regex/) {
										$result=!$negate;
									} else {
										$result=$negate;
									}
									if(!$result) {
										if(DEBUG>= 1) { LogMsg("Match: No Match! (negate is $negate)"); }
										last;
									}
									$count++;
								}
								#$result=Match($element,$match,$negate);
							}
							#if($result && !$negate || !$result && $negate) {}
							if($result) {
								# Yay - one of the alternatives matched
								if(DEBUG>= 1) { LogMsg("BREAKING out (result is $result, negate is $negate)"); }
								last;
							}
						}
						if($result) {
							$ObjectiveMatch=1;
						} else {
							$ObjectiveMatch=0;
							next FilterLOOP;
						}
					}
				}
			} else {
				# No token in this event that matches? Drop it.
				$ObjectiveMatch=0;
			}
		}
		if($ObjectiveMatch) {
			if ($Criticality[$objectivecount] < 0) {
				if(DEBUG>= 2) { LogMsg("BREAKING out, event should be ignored"); }
				return;
			}
			$EventMatched=1;
			# if you catch any new files, watch them too
			#if ($tokens{"event"} =~ /^open,/ && $tokens{"name"} !~ /\/$tokens{"watch"}$/) {
				#LogMsg(" TOKENS: " . $tokens{"watch"} . ":" . $tokens{"name"});
			#}
			if ($tokens{"filterkey"} eq "dir" && $tokens{"name"} !~ /\/$tokens{"watch"}$/) {
				if ($tokens{"name"} && not defined $WatchExtra{$tokens{"name"}}) {
					ExtraWatch($tokens{"name"}, WATCH);
					$WatchExtra{$tokens{"name"}}=1;
				}
			}

			if($USECRITICALITY) {
				$tcrit=$Criticality[$objectivecount];
				if($tcrit > $EventCrit) {
					$EventCrit=$tcrit;
				}
			} else {
				$EventCrit=$Criticality[$objectivecount];
				last;
			}
		}
	}

	if(!$EventMatched) {
		# Ok, go to the next objective.
		# We're not interested in this one.
		return;
	}

	# Craft our tokens into a single string.
	$sendstring="criticality,$EventCrit";

	foreach $key (@tokenlist) {
		$sendstring .= "	$key,".$tokens{$key};
	}

	# Send this out.
	if(DEBUG) { LogMsg("OUTPUT: " . $sendstring); }

	SendEvent("$FQDN	LinuxKAudit","$sendstring\n");
}

sub SendEvent() {
	my($prefix,$string)=@_;
	my $original_string="$prefix\t$string";

	$success=0;
	if (defined $OUTPUTFILE) {
		print $OUTPUTFILE "$prefix\t$string";
		$success=1;
	}
	if (defined $OUTPUTNET{"Socket"}) {
		if($OUTPUTNET{"Port"} == 514) {
			# Syslog = special
			# Prepend some syslog gumf. REDRED TODO
			$sdate=strftime("%b %e %H:%M:%S",localtime);
			$prefix="<$SYSLOGDEST> $sdate $FQDN LinuxKAudit";
			#$prefix="<$SYSLOGDEST> LinuxKAudit";
		}
		send($OUTPUTNET{"Socket"}, "$prefix\t$string", 0, $OUTPUTNET{"PortAddr"});
		$success=1;
	}

	if($success && $webpid) {
		# Cache the last 100 events, for the web.
		$WebCacheSize=@WebCache;
		if($WebCacheSize > 99) {
			shift(@WebCache)
		}
		push(@WebCache,$original_string);
	}
}

sub CloseOutputs() {
	if(defined $OUTPUTFILE) {
		close($OUTPUTFILE);
	}
	undef $OUTPUTFILE;

	if(defined $OUTPUTNET{"Socket"}) {
		close($OUTPUTNET{"Socket"});
	}
	delete $OUTPUTNET{"Socket"};
}

sub LoadConfig() {
	@Criticality=();
	@Filters=();
	@FilterTypes=();
	@Output=();
	%WatchVal=();
	%WatchExtra=();
	$WatchCount=0;

	$rc=open(CONFIG,"$CONFIGFILE");
	if(!$rc) {
		LogMsg("Cannot open Snare Configuration File! Exiting.");
		return(0);
	}
	$section="Unknown";
	$ObjectiveCount=0;

	LINE: while($line=<CONFIG>) {
		chomp($line);
		if($line =~ /^[ \t]*#/ || $line =~ /^[ \t#]*$/) {
			# Ignore
			next;
		} elsif ($line =~ /^[ \t]*\[[a-zA-Z]+\]/) {
			($null,$section)=split(/[\[\]]/,$line);
			$section =~ tr/a-z/A-Z/;
		} else {
			$line =~ s/^[ \t]+//;
			$line =~ s/[ \t]+/\t/g;
			$line =~ s/[ \t]+$//;

			if(DEBUG) { LogMsg("Config File: Loaded line $line\n"); }

			if($section eq "REMOTE") {
				($key,$val)=split(/=/,$line);
				$key =~ tr/a-z/A-Z/;
				if($key eq "ALLOW") {
					$WEBSERVERALLOW=$val;
				} elsif($key eq "LISTEN_PORT") {
					$WEBSERVERPORT=$val;
				} elsif($key eq "ACCESSKEY") {
					$ACCESSKEY=$val;
				} elsif($key eq "RESTRICT_IP") {
					$RESTRICT_IP=$val;
				}
			} elsif($section eq "OUTPUT") {
				# Add this into our output file array
				$rc=OpenOutput($line);
			} elsif($section eq "CONFIG") {
				($key,$val)=split(/=/,$line);
				$key =~ tr/a-z/A-Z/;
				if($key eq "USE_CRITICALITY") {
					$USECRITICALITY=$val;
				} elsif($key eq "USE_REGEX") {
					$REGEX=$val;
				} elsif($key eq "USE_WATCH") {
					$USEWATCH=$val;
				} elsif($key eq "CLIENTNAME") {
					$FQDN=$val;
				} elsif($key eq "SET_AUDIT") {
					$MANAGEAUDIT=$val;
				} elsif($key eq "SYSLOG_FACILITY") {
					$SFACILITY=$val;
					$SFACILITY =~ tr/a-z/A-Z/;
				} elsif($key eq "SYSLOG_PRIORITY") {
					$SPRIORITY=$val;
					$SPRIORITY =~ tr/a-z/A-Z/;
				}
			} elsif($section eq "WATCH") {
				@elements=split(/\t/,$line);
				
				$path="";
				$new=0;
				foreach $element (@elements) {
					($key,$check,$val)=split(/([\!~]?=)/,$element);
					if($key eq "path") {
						# paths must be absolute
						if($val !~ /^\//) {
							LogMsg("Watch only allows absolute paths, ignoring [$val]");
							next LINE;
						}
						$path = resolve_path($val);
						# make sure the path is valid and not the root dir (/)
						if ($path !~ /^\/$/) {
							# check if we should watch or ignore a given path
							if ($check eq "=" || $check eq "~=") {
								# watches need a valid path
								if (-e $path) {
									$WatchVal{$WatchCount}{"path"} = $path;
									if ($check eq "~=") {
										$WatchVal{$WatchCount}{"flag"} = NEWWATCH;
									} else {
										$WatchVal{$WatchCount}{"flag"} = WATCH;
									}
								} else {
									LogMsg("WATCH requires a valid path, ignoring [$val]");
									#$WatchVal{$path} = WATCH;
									next LINE;
								}
							} else {
								# ignore can be any value, just strip out backslash and double quotes
								$path =~ s/\\"//g;
								$WatchVal{$WatchCount}{"path"} = $path;
								$WatchVal{$WatchCount}{"flag"} = IGNORE;
							}
						} else {
							LogMsg("WATCH requires a valid path that is not /, ignoring [$val]");
							next LINE;
						}
					} else {
						LogMsg("Invalid keyword \"$key\" found in [Watch]");
					}
				}
				$WatchCount++;
			} elsif($section eq "OBJECTIVES") {
				# Default value.
				$TEMPObjectiveCount = $ObjectiveCount;
				$Criticality[$TEMPObjectiveCount]=0;

				@elements=split(/\t/,$line);
				%Config=();
				%OPEN_EVENTS=('open' => 1,'creat' => 1,'link' => 1,'symlink' => 1,'truncate' => 1,'ftruncate' => 1,'mknod' => 1,'rename' => 1,'truncate64' => 1,'ftruncate64' => 1,'access' => 1);
				
				foreach $element (@elements) {
					($key,$check,$val)=split(/(\!*=)/,$element);
					if($key eq "watch") {
						# ignore, these no longer belong here
						LogMsg("Please move all watches to [watch]");
						next;
					}
					if($key eq "criticality" && $val ne "") {
						$Criticality[$TEMPObjectiveCount]=$val;
						next;
					}

					@components=();
					@tcomponents=split(/,/,$val);
					while(@tcomponents) {
						#take a copy of tcomponents in case the brackets are invalid
						@temp_tc = @tcomponents;
						$component=shift(@tcomponents);
						if($component =~ /^\(/) {
					#LogMsg("DMM: check comp (:$component");
							while($component !~ /\)$/ && @tcomponents) {
								$component .= "," . shift(@tcomponents);
							}
							if ($component !~ /\)$/) {
								#bad use of brackets
								LogMsg("WARNING: Removing unmatched bracket");
								@tcomponents = @temp_tc;
								$component=shift(@tcomponents);
							}
							$component =~ s/^\(//;
							$component =~ s/\)$//;
						}
						push(@components,$component);
					}
					
					$ElementCount=0;
					$keymatch=keys(%{$Filters{$TEMPObjectiveCount}{$key}});
					foreach $component (@components) {
						$AlternativeCount=0;
						@alternatives = split(/,/,$component);
						foreach $alternative (@alternatives) {
							#LogMsg("Adding $alternative to $TEMPObjectiveCount:$key:$keymatch:$ElementCount:$AlternativeCount");
							$Filters{$TEMPObjectiveCount}{$key}{$keymatch}{$ElementCount}{$AlternativeCount}=$alternative;
							if($check eq "=") {
								# If the user has specified a positive match, and
								# Has identified an explicit event, then mark it to be turned on.
								if($key eq "event") {
									if (!($USEWATCH && $OPEN_EVENTS{$alternative} == 1)) {
										# put the events in a temporary list
										$temp_EventsON{$alternative}=SUCCESS + FAILURE;
									}
								} elsif($key eq "return") {
									# if we are only looking for a specific return code,
									# only enable the necessary auditing
									if ($alternative =~ /,no/) {
										foreach $eventid (keys(%temp_EventsON)) {
											$temp_EventsON{$eventid}= $temp_EventsON{$eventid} | FAILURE;
										}
									} elsif ($alternative =~ /,yes/) {
										foreach $eventid (keys(%temp_EventsON)) {
											$temp_EventsON{$eventid}= $temp_EventsON{$eventid} | SUCCESS;
										}
									}
								}
								$FilterTypes{$TEMPObjectiveCount}{$key}{$keymatch}=0;
							} else {
								# Negate this match
								$FilterTypes{$TEMPObjectiveCount}{$key}{$keymatch}=1;
							}
							$AlternativeCount++;
						}
						$ElementCount++;
					}
				}
				foreach $eventid (keys(%temp_EventsON)) {
					$EventsON{$eventid} = $temp_EventsON{$eventid};
					delete $temp_EventsON{$eventid};
				}

				$ObjectiveCount++;
			}
		}
	}
	# remove WatchVal if we are not using watches
	undef %WatchVal if (!$USEWATCH);

	$FilterCount = $ObjectiveCount;
	# Work out syslog facility/priority number:
	# $SFACILITY $SPRIORITY (all upper)
	$sfac{"KERNEL"}=0;	$sfac{"USER"}=1;	$sfac{"MAIL"}=2;
	$sfac{"DAEMON"}=3;	$sfac{"AUTH"}=4;	$sfac{"SYSLOG"}=5;
	$sfac{"LPR"}=6;		$sfac{"NEWS"}=7;	$sfac{"UUCP"}=8;
	$sfac{"CRON"}=9;	$sfac{"AUTHPRIV"}=10;	$sfac{"FTP"}=11;
	$sfac{"LOCAL0"}=16;	$sfac{"LOCAL1"}=17;	$sfac{"LOCAL2"}=18;
	$sfac{"LOCAL3"}=19;	$sfac{"LOCAL4"}=20;	$sfac{"LOCAL5"}=21;
	$sfac{"LOCAL6"}=22;
	$spri{"EMERGENCY"}=0;	$spri{"ALERT"}=1;	$spri{"CRITICAL"}=2;
	$spri{"ERROR"}=3;	$spri{"WARNING"}=4;	$spri{"NOTICE"}=5;
	$spri{"INFORMATION"}=6;	$spri{"DEBUG"}=7;
	
	$SYSLOGDEST=($sfac{$SFACILITY} << 3) | $spri{$SPRIORITY};

	return(1);
}

sub SetAudit() {
	if($MANAGEAUDIT == 1) {
		# temporarily disable audit while we modify the values
		`/sbin/auditctl -e 0`;
		# Clear all audit events
		`/sbin/auditctl -D`;
		HandleWatch(\%WatchVal);
		if(%EventsON) {
			$list_sf="";
			$list_s="";
			$list_f="";

			foreach $eventid (keys(%EventsON)) {
				if (defined($nsyscall{$eventid}) || defined($syscall{$eventid})) {
					if ($EventsON{$eventid} == (SUCCESS + FAILURE)) {
						if(DEBUG) { LogMsg("Turning on event $eventid"); }
						$list_sf .= " -S $eventid";
					} elsif ($EventsON{$eventid} == SUCCESS) {
						if(DEBUG) { LogMsg("Turning on event $eventid (SUCCESS only)"); }
						$list_s .= " -S $eventid";
					} elsif ($EventsON{$eventid} == FAILURE) {
						if(DEBUG) { LogMsg("Turning on event $eventid (FAILURE only)"); }
						$list_f .= " -S $eventid";
					} else {
						if(DEBUG) { LogMsg("fail $eventid"); }
					}
				} else {
					if(DEBUG) { LogMsg("Cannot find syscall: $eventid"); }
				}
			}
			if ($list_sf ne "") {
				if (DEBUG) { LogMsg("/sbin/auditctl -a entry,always $list_sf"); }
				`/sbin/auditctl -a entry,always $list_sf`;
			}
			if ($list_s ne "") {
				`/sbin/auditctl -a entry,always $list_s -F success=1`;
			}
			if ($list_f ne "") {
				`/sbin/auditctl -a exit,always $list_f -F success=0`;
			}
		}
		# if these events are not explicitly turned on, then block them out
		# since they can be quite noisy using watches
		if (not defined $EventsON{"lstat64"} && not defined $EventsON{"stat64"} && not defined $EventsON{"access"}) {
			`/sbin/auditctl -A entry,never -S lstat64 -S stat64 -S access`;
		}
		`/sbin/auditctl -e 1`;
	}
}

sub StartWeb() {
	$webpid=0;
	if($WEBSERVERALLOW) {
		if(-f $WEBSERVER) {
			require $WEBSERVER;

			$webpid=open(WEBPROCESS,"|-");
			if($webpid==0) {
				# Child process
				WebServerSetup($WEBSERVERPORT,$ACCESSKEY,$RESTRICT_IP);
				exit;
			}
			# No buffering!
			$old_fh = select(WEBPROCESS);
			$| = 1;
			select($old_fh);
			LogMsg("Snare Configuration Web Server active and running.");
		}
	}
}

sub Match {
	my($string,$match,$negate)=@_;

	# Commas are special.
	@strings=split(/,/,$string);
	@matches=split(/,/,$match);
	$count=0;
	foreach $tmatch (@matches) {
		if ($REGEX == 1) {
			$regex=$tmatch;
			$regex =~ s/[^\\]\//\\\//g;
		} else {
			$regex=wildcard($tmatch);
		}
		if(DEBUG>=1) { LogMsg("Match: $regex against " . $strings[$count]); }
		if($strings[$count] =~ /$regex/) {
			$match=!$negate;
		} else {
			$match=$negate;
		}
		if(!$match) {
			if(DEBUG>= 1) { LogMsg("Match: No Match! (negate is $negate)"); }
			last;
		}
		$count++;
	}
	return($match);
}

sub OpenOutput() {
	my $line = shift;
	($type,$destination,$opt,$opt2)=split(/[=:]/,$line);
	if($type eq "file" && not defined $OUTPUTFILE) {
		if($destination eq "stdout") {
			open($OUTPUTFILE,">-");
		} else {
			open($OUTPUTFILE,">>$destination");
			# hot file handle to stop unwanted buffering
			select((select($OUTPUTFILE), $|=1)[0]);
		}
		if(defined $OUTPUTFILE) {
			return(1);
		}
		return(0);
	} elsif($type eq "network") {
		# Destination will be an ip or dns name
		# opt will be the port
		if (defined $OUTPUTNET{"Socket"}) {
			return(0);
		}
		$Port=6161;
		if($opt > 0 && $opt < 65536) {
			$Port=$opt;
		}
		if($destination =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
			# IP Address
			$ipaddr = inet_aton($destination);
		} else {
			$ipaddr = gethostbyname($destination);
		}
		if(!$ipaddr) {
			return(0);
		}
		$portaddr = sockaddr_in($Port, $ipaddr);
		if(!$portaddr) {
			return(0);
		}
		socket($tsocket, PF_INET, SOCK_DGRAM, getprotobyname("udp")) or return(0);
		
		$OUTPUTNET{"Socket"}=$tsocket;
		$OUTPUTNET{"PortAddr"}=$portaddr;
		$OUTPUTNET{"Port"}=$Port;
		$OUTPUTNET{"Line"}=$line;
		return(1);
	}
	return(0);
}

#
# wildcard to regex conversion
#
sub wildcard ($) {
	my $pat = shift or return(".*");
	my $re = quotemeta $pat;
	$re =~ s/\\\*/(.*)/g;
	return("^" . $re . "\$");
}

sub resolve_path {
	my($path)=@_;
	my(@pathparts)=split(/\//,$path);
	my(@newpath)=();

	foreach $part (@pathparts) {
		if($part eq "..") {
			if(@newpath) {
				pop(@newpath);
			}
		} elsif($part eq ".") {
			# Do nothing
		} elsif($part eq "") {
			# Do nothing
		} else {
			# strip out double quotes
			$part =~ s/"//g; 
			push(@newpath,$part);
		}
	}
	$resolvedpath="";
	if($path =~ /^\//) {
		$resolvedpath = "/";
	}
	$resolvedpath .= join("/",@newpath);

	return($resolvedpath);
}

sub getuname {
	my($name)=@_;
	if($unamecache{"$name"} eq "-") {
		return(-1);
	} elsif($unamecache{"$name"}) {
		return($unamecache{"$name"});
	} else {
		$id=getpwnam($name);
		if($id) {
			$unamecache{"$name"}=$id;
			return($id);
		} else {
			$unamecache{"$name"}="-";
			return(-1);
		}
	}
}

sub getuid {
	my($id,$sleep)=@_;
	if($uidcache{"$id"} eq "-") {
		return(0);
	} elsif($uidcache{"$id"}) {
		return($uidcache{"$id"});
	} else {
		$name=getpwuid($id);
		if($sleep == 1 && !$name) {
			sleep(1);
			$name=getpwuid($id);
		}
		if($name) {
			$uidcache{"$id"}=$name;
			return($name);
		} else {
			$uidcache{"$id"}="-";
			return(0);
		}
	}
}

sub getgid {
	my($id)=@_;
	if($gidcache{"$id"} eq "-") {
		return(0);
	} elsif($gidcache{"$id"}) {
		return($gidcache{"$id"});
	} else {
		$name=getgrgid($id);
		if($name) {
			$gidcache{"$id"}=$name;
			return($name);
		} else {
			$gidcache{"$id"}="-";
			return(0);
		}
	}
}

sub Handler {
	local($sig)=@_;
	LogMsg("SIG$sig Signal Received");
	if(DEBUG) { LogMsg("SIG$sig Signal Received"); }
	if ($sig ne "PIPE") {
		$CONTINUE=0;
	}
}

sub Restart {
	local($sig)=@_;
	if(DEBUG) { LogMsg("SIG$sig Signal Received in restart routine"); }
	#$CONTINUE=0;
	#if($webpid) {
	#	# send sigusr1, which tells the web server thread to terminate cleanly.
	#	kill(10,$webpid);
	#}

	# Flush our buffers
	CloseOutputs();
	sleep(1);

	%event=();
	%Filters=();

	LoadConfig();
	StartWeb();
	# Set up our audit configuration based on our objectives.
	SetAudit();
	# Good to go.
}

sub DumpEvents {
	local($sig)=@_;
	if($webpid) {
		$EventCount=@WebCache;
		if($EventCount) {
			foreach $event (reverse(@WebCache)) {
				chomp($event);
				print WEBPROCESS "$event\n";
			}
			print WEBPROCESS "END\n";
		}
	}
}


sub LogMsg {
	my($string)=@_;
	if(!$string) {
		return;
	}
	setlogsock('unix');
	$rc=openlog("SnareDispatcher",'','user');
	if($rc) {
		syslog('info',"[$$]$string");
	}
	closelog();
	print "DEBUG: $string\n";
}

sub ExtraWatch {
	my ($path,$flag) = @_;
	if (DEBUG >= 2) { LogMsg("Handling watch for $path"); }
	$tempfile = "/tmp/snarewatch.txt";
	`/usr/bin/find "$path" -type d -printf "-i -w %p -k dir\n" -or -type f -printf "-i -w %p\n" >> $tempfile`;
	chmod 0600,$tempfile;
	`/sbin/auditctl -R $tempfile` if (-e $tempfile);
	unlink ($tempfile);
}

sub HandleWatch {
	my(%watch) = %{$_[0]};
	#loop through the watch values
	$tempfile = "/tmp/snarewatch.txt";
	# make sure it is empty
	unlink ($tempfile);
	my $new=0;
	for ($wcount=0; $wcount < $WatchCount;$wcount++) {
		#first, check if we need to watch for any new values, if not, it makes things easy
		$new=1 if ($watch{$wcount}{"flag"} == NEWWATCH);
	}
	if (!$new) {
		for ($wcount=0; $wcount < $WatchCount;$wcount++) {
			$path = $watch{$wcount}{"path"};
			if (DEBUG >= 2) { LogMsg("Handling watch for " . $watch{$wcount}{"path"}); }
			if ($watch{$wcount}{"flag"} == WATCH) {
				`/usr/bin/find "$path" -type f -printf "-i -w %p\n" >> $tempfile`;
			} else {
				if (-d $path) {
					`/usr/bin/find "$path" -printf "-i -W %p\n" >> $tempfile`;
				} else {
					`echo "-i -W $path" >> $tempfile`;
				}
			}
		}
		chmod 0600,$tempfile;
		`/sbin/auditctl -R $tempfile` if (-e $tempfile);
		unlink ($tempfile);
	} else {
		for ($wcount=0; $wcount < $WatchCount;$wcount++) {
			$path = $watch{$wcount}{"path"};
			if (DEBUG >= 2) { LogMsg("Handling watch for " . $watch{$wcount}{"path"}); }
			if ($watch{$wcount}{"flag"} == NEWWATCH) {
				`/usr/bin/find "$path" -type d -printf "-i -w %p -k dir\n" -or -type f -printf "-i -w %p\n" >> $tempfile`;
			} elsif ($watch{$wcount}{"flag"} == WATCH) {
				#Just in case the parent dir in part of NEWWATCH
				if (-d $path) {
					$audit=`auditctl -l | grep "path=$path, filterkey=dir"`;
					if ($audit ne '') {
						# NEWWATCH zone, will need to use snareignore on the directories
						`/usr/bin/find "$path" -type d -printf "-i -W %p\n-i -w %p -k snareignore\n" >> $tempfile`;
					}
				}
				`/usr/bin/find "$path" -type f -printf "-i -w %p\n" >> $tempfile`;
			} else {
				#IGNORE
				if (-d $path) {
					$audit=`auditctl -l | grep "path=$path, filterkey=dir"`;
					if ($audit ne '') {
						# if the dir being ignored is in a NEWWATCH zone, need to use snareignore
						`/usr/bin/find "$path" -printf "-i -W %p\n-i -w %p -k snareignore\n" >> $tempfile`;
					} else {
						# otherwise, just strip out the file watches
						`/usr/bin/find "$path" -type f -printf "-i -W %p\n" >> $tempfile`;
					}
				} else {
					# ignoring a file, file might not exist
					# check that the basedir is not being audited
					$basedir = dirname($path);
					$audit=`auditctl -l | grep "path=$basedir, filterkey=dir"`;
					if ($audit ne '') {
						# NEWWATCH zone, will need to use snareignore
						`echo "-i -W $path\n-i -w $path -k snareignore" >> $tempfile`;
					} else {
						# otherwise, just strip out the file watch
						`echo "-i -W $path" >> $tempfile`;
					}
				}
			}
			chmod 0600,$tempfile;
			`/sbin/auditctl -R $tempfile` if (-e $tempfile);
			unlink ($tempfile);
		}
	}
}
